#ifndef DICTIONARY_H
#define DICTIONARY_H

#include "list.h"
#include "listiterator.h"

template <class K, class V> class association
{
public:
			// Constructor
			association(K initialKey, V initialValue);

		void	operator = (const association<K,V> &);
		void	operator = (V val);

	// accessing the pieces
	K	Key() const {return fKey;};
	V	Value() const {return fValue;};

	// Public for easy access
	V	fValue;

protected:
	// This wants to be const, but then operator= won't work
	K	fKey;

private:

};


template <class K, class V>
association<K, V>::association(K aKey, V aValue)
	: fKey(aKey)
{
	fValue = aValue;
}

template <class K, class V>
void
association<K,V>::operator=(const association<K, V> & rhs)
{
	fKey = rhs.fKey;
	fValue = rhs.fValue;
}

template <class K, class V>
void
association<K,V>::operator=(V aValue)
{
	fValue = aValue;
}

//=====================================================
// template interface: dictionary
//
// This is a very basic dictionary that stores its value
// using a list of associations.  There is no attempt at
// ordering that takes place.
//=====================================================

template <class K, class V> class dictionary
{
public:
				dictionary();
				dictionary(V defaultValue);
				dictionary(const dictionary<K,V> &);
				
	// Dictionary protocol
	virtual V &	operator [] (K key);
	virtual void	DeleteAll();
	virtual void	Remove(K key);
	virtual void	SetDefault(V value);

	virtual bool	Includes(K key);
	virtual bool	IsEmpty();
		
protected:
	list<association<K, V> *>	fData;
	V		fDefaultValue;
	
	friend class dictionaryIterator<K,V>;
//	friend class table<K,V>;
	
	association<K,V> *	FindAssociation(K key);
	
private:
};


//=====================================================
// template implementation: dictionary
//=====================================================

template <class K, class V>
dictionary<K,V>::dictionary()
{
}

template <class K, class V>
dictionary<K,V>::dictionary(V defaultValue)
	: fDefaultValue(defaultValue)
{
}

template <class K, class V>
dictionary<K,V>::dictionary(const dictionary<K,V> &other)
	: fDefaultValue(other.fDefaultValue),
	fData(other.fData)
{
}

template <class K, class V>
void
dictionary<K,V>::DeleteAll()
{
	fData.DeleteAll();
}

template <class K, class V>
void
dictionary<K,V>::SetDefault(V value)
{
	fDefaultValue = value;
}

template <class K, class V>
bool
dictionary<K,V>::IsEmpty()
{
	return fData.IsEmpty();
}

template <class K, class V>
bool
dictionary<K,V>::Includes(K key)
{
	return FindAssociation(key) != 0;
}

template <class K, class V>
V &
dictionary<K,V>::operator[] (K key)
{
	association<K,V> * newAssoc = FindAssociation(key);
	
	if (!newAssoc)
	{
		newAssoc = new association<K,V>(key, fDefaultValue);
		assert(newAssoc != 0);
		fData.Add(newAssoc);
	}

	return newAssoc->fValue;
}

template <class K, class V>
void
dictionary<K,V>::Remove(K key)
{
	listIterator<association<K,V> *> itr(fData);
	for (itr.init(); !itr.done(); itr.next())
		if (itr()->Key() == key)
			itr.RemoveCurrent();
}

template <class K, class V>
association<K,V> *
dictionary<K,V>::FindAssociation(K key)
{
	listIterator<association<K,V> *> itr(fData);
	for (itr.init(); !itr.done(); itr.next())
		if (itr()->Key() == key)
			return itr();

	return 0;
}



//=========================================================
// Template Interface: dictionaryIterator
//
// This iterator will traverse a dictionary touching each 
// entry in turn.
//=========================================================

template <class K, class V>
class dictionaryIterator : public listIterator<association<K,V> *>
{
public:
			dictionaryIterator(dictionary<K,V> &dict);
			
protected:
private:
};


template <class K, class V>
dictionaryIterator<K,V>::dictionaryIterator(dictionary<K,V> &dict)
	: listIterator<association<K,V> *>(dict.fData)
{
}

#endif