#ifndef QUEUERINGBUFFER_H
#define QUEUERINGBUFFER_H

#include "queue.h"
#include "vector.h"

//========================================================
// Class: queueRingBuffer
//
// This is an implementation of a ringbuffer which uses
// the queue protocol for access.
//========================================================

template <class AType> class queueRingBuffer : public queue<AType>
{
public:
		queueRingBuffer(const unsigned long size);
		queueRingBuffer(const queueRingBuffer& other);
		
	virtual unsigned long	Capacity() const;	// How many can it hold
	virtual unsigned long	Size() const;		// How many currently held
	
	// Implementation of queue protocol
	virtual	void	MakeEmpty();
	virtual AType	Dequeue();
	virtual void	Enqueue(const AType val);
	virtual AType	Front();
	virtual int		IsEmpty() const;
	

protected:
	vector<AType>		fData;		// The sequence that holds the data
	const unsigned long	fMaxSize;	// Max num items in queue
	unsigned long		fSize;		// How many items in queue
	unsigned long		fNextSlot;	// Where next item will be enqueued
	unsigned long		fFront;		// Where Front and Dequeue will get item from
};


//===================================================
// Implementation: queueRingBuffer
//===================================================

template <class AType> 
queueRingBuffer<AType>::queueRingBuffer(const unsigned long size)
	: fMaxSize(size),
	fData(size),
	fSize(0)
{
	// Initialize sequence to being empty
	DeleteAll();
}

template <class AType> 
queueRingBuffer<AType>::queueRingBuffer(const queueRingBuffer& other)
	: fMaxSize(other.fMaxSize),
	fData(other.fData),
	fNextSlot(other.fNextSlot),
	fFront(other.fFront),
	fSize(other.fSize)
{
}

template <class AType> 
void 
queueRingBuffer<AType>::MakeEmpty()
{
	fNextSlot = 0;
	fFront = 0;
	fSize = 0;
}

template <class AType> 
int
queueRingBuffer<AType>::IsEmpty() const
{
	return (fNextSlot == fFront);
}


template <class AType>
unsigned long
queueRingBuffer<AType>::Capacity() const
{
	return fMaxSize;	
}

template <class AType> 
void
queueRingBuffer<AType>::Enqueue(const AType val)
{
	fData[fNextSlot] = val;
	fNextSlot++;
	if (fNextSlot >= fMaxSize)
		fNextSlot = 0;
		
	if (fSize < fMaxSize)
		fSize++;
}

template <class AType> 
AType
queueRingBuffer<AType>::Dequeue()
{
	if (IsEmpty())
		return fData[fFront];
		
	unsigned long dataloc = fFront;
	fFront++;
	if (fFront >= fMaxSize)
		fFront = 0;
	
	if (fSize > 0)
		fSize--;

	return fData[dataloc];
}

template <class AType> 
AType
queueRingBuffer<AType>::Front()
{
	return fData[fFront];
}

#endif
