#ifndef SEARCHTREE_H
#define SEARCHTREE_H

#include "binarytree.h"
#include "treetraverse.h"

// Forward declarations
template <class T> class searchTreeIterator;
template <class T> class searchTree;

//==============================================================
// Template Interface: searchTree<T>
//
// T - The type of values stored in the tree.  This type must
// implement the operator <, operator ==
//==============================================================
template <class T> 
class searchTree : public binaryTree<T>
{
public:
				searchTree();
				searchTree(const searchTree<T> &);
				~searchTree();
				
	virtual bool	Add(T value);
	virtual void	Remove(const T value);
			bool	Includes(const T value) const;

	virtual binaryNode<T> *	Find(const T value) const;
	
protected:
	friend class searchTreeIterator<T>;
	
	binaryNode<T> * RemoveTop(binaryNode<T> *);
	
private:
};


//------------------------------------------------------------
//------------------------------------------------------------
template <class T>
searchTree<T>::searchTree()
{
}

//------------------------------------------------------------
//------------------------------------------------------------
template <class T>
searchTree<T>::searchTree(const searchTree<T> &rhs)
{
	searchTreeIterator<T> itr(rhs);
	for (itr.init(); !itr.done(); itr.next())
		Add(itr());
}

//------------------------------------------------------------
//------------------------------------------------------------
template <class T>
searchTree<T>::~searchTree()
{
}

//------------------------------------------------------------
//------------------------------------------------------------
template <class T>
bool
searchTree<T>::Add(T value)
{
	if (fRoot)
	{
		binaryNode<T> *current = fRoot;
		binaryNode<T> *child;
		
		while (current)
		{
			if (current->fValue < value)
			{
				child = current->Right();
				if (!child)
				{
					current->SetRight(new binaryNode<T>(value));
					return true;
				}
			} else
			{
				child = current->Left();
				if (!child)
				{
					current->SetLeft(new binaryNode<T>(value));
					return true;
				}
			}
			
			current = child;
		}
	} else
		fRoot = new binaryNode<T>(value);
		
	// Always return true because adding never fails
	// unless we run out of memory.
	return true;
}


//------------------------------------------------------------
//------------------------------------------------------------
template <class T>
binaryNode<T> *
searchTree<T>::Find(const T value) const
{
	if (fRoot)
	{
		binaryNode<T> *current = fRoot;
		while(current)
		{
			if (current->fValue == value)
				return current;
			if (current->fValue < value)
				current = current->Right();
			else
				current = current->Left();
		}
	}

	// If we got to here, then the value is not
	// in the tree.
	return 0;
}

//------------------------------------------------------------
//------------------------------------------------------------
template <class T>
bool
searchTree<T>::Includes(const T value) const
{
	if (Find(value))
		return true;
		
	return false;
}

template <class T>
void
searchTree<T>::Remove(const T value)
{
	// If we don't have a root, then just return
	if (!fRoot)
		return ;

	// If it's the root node we're after, then we need 
	// to remove it in a special way.
	if (fRoot->fValue == value)
	{
		binaryNode<T> *newRoot = RemoveTop(fRoot);
		delete fRoot;
		fRoot = newRoot;
		return ;
	}
	
	// It's not the root we're removing, so find the right
	// node, and remove it.
	binaryNode<T> *parent = fRoot;
	while (parent != 0)
	{
		binaryNode<T> *child;
		if (parent->fValue < value)
		{
			// Search right sub-tree
			child = parent->Right();
			if (child &&(child->fValue == value))
			{
				parent->SetRight(RemoveTop(child));
				delete child;
				child = 0;
			}
		} else
		{
			// Search left sub-tree
			child = parent->Left();
			if (child && (child->fValue == value))
			{
				parent->SetLeft(RemoveTop(child));
				delete child;
				child = 0;
			}
		}
		parent = child;
	}
}

template <class T>
binaryNode<T> *
searchTree<T>::RemoveTop(binaryNode<T> *topNode)
{
	binaryNode<T> *left = topNode->Left();
	binaryNode<T> *right = topNode->Right();
	
	// No left node
	if (0 == left)
		return right;
	
	// No right node
	if (0 == right)
		return left;
	
	// Right node has no left node
	binaryNode<T> *n = right->Left();
	if (0 == n)
	{
		right->SetLeft(left);
		return right;
	}
	
	// slide down left tree
	binaryNode<T> *parent = right;
	while (n->Left() != 0)
	{
		parent = n;
		n = n->Left();
	}
	
	parent->SetLeft(n->Right());
	n->SetLeft(left);
	n->SetRight(right);
	
	return n;
}



//==============================================================
// Template Interface: searchTreeIterator<T>
//
// A simple iterator for a searchTree.  You could simply use the
// root node of the search tree and use any of the pre,post,inorder
// traversers, but then you would be exposing the root node of the
// tree, which may not be desirable.
//
// This iterator will do an inorder iteration of the tree.
//==============================================================

template <class T>
class searchTreeIterator : public inorderTraversal<T>
{
public:
				searchTreeIterator(searchTree<T> &);
				
	virtual bool	init();
	
protected:
	searchTree<T>	& fTree;
	
private:
};

template <class T>
searchTreeIterator<T>::searchTreeIterator(searchTree<T> &tree)
	: inorderTraversal<T>(tree.fRoot),
	fTree(tree)
{	
}

template <class T>
bool
searchTreeIterator<T>::init()
{
	SetRoot(fTree.fRoot);
	return inorderTraversal<T>::init();
}


#endif
