#ifndef SPLAYTREE_H
#define SPLAYTREE_H

#include "binarynode.h"

//==================================================================
// Template Interface: splayTree<T>
//
// A Splay Tree is a nifty type of tree that gives good
// performance on mid-sized data sets (65535 elements).  It is fast
// and relatively easy to implement.
//
// class T: Must have a default constructor, copy constructor,
//			operator <, and operator ==
//
// The splayTree is constructed without parameters and cannot be
// copied, or assigned.			
//==================================================================

template <class T>
class splayTree
{
public:
			splayTree();
	virtual ~splayTree();

	bool	Insert(const T &);
	bool	Add(const T &);
	bool	Remove(const T &);
	
	bool	Includes(const T & X) {Splay(X, fRootNode);
				return fRootNode != NilNode && fRootNode->fValue == X;}

	const T & GetRoot() const {return fRootNode->fValue;};
	
	void	MakeEmpty() {FreeTree(fRootNode); fRootNode = NilNode;}
	
	bool	IsEmpty() const { return fRootNode == NilNode;}
	
protected:
	binaryNode<T>	*NilNode;
	binaryNode<T>	*fRootNode;
	
	void FreeTree(binaryNode<T> *node);
	void Splay(const T & val, binaryNode<T> * &node);
	
	// Copying a splayTree is not allowed
	splayTree(const splayTree<T> &);
	const splayTree & operator=(const splayTree<T> &);
	
private:
};


//-------------------------------------------------------------------
// Template Implementation: splayTree<T>
//-------------------------------------------------------------------

template <class T>
splayTree<T>::splayTree()
{
	NilNode = new binaryNode<T>;
	NilNode->SetLeft(NilNode);
	NilNode->SetRight(NilNode);
	fRootNode = NilNode;
};

template <class T>
splayTree<T>::~splayTree()
{
	if (fRootNode != NilNode)
	{
		FreeTree (fRootNode);
	}
	delete NilNode;
}

template <class T>
bool
splayTree<T>::Add(const T & val)
{
	static binaryNode<T> *newNode = 0;
	
	if (newNode == 0)
		newNode = new binaryNode<T>(val);

	if (fRootNode == NilNode)
	{
		newNode->SetLeft(NilNode);
		newNode->SetRight(NilNode);
		fRootNode = newNode;
	} else
	{
		Splay(val, fRootNode);
		if (val < fRootNode->fValue)
		{
			newNode->SetLeft(fRootNode->Left());
			newNode->SetRight(fRootNode);
			fRootNode->SetLeft(NilNode);
			fRootNode = newNode;
		} else if (fRootNode->fValue < val)
		{
			newNode->SetRight(fRootNode->Right());
			newNode->SetLeft(fRootNode);
			fRootNode->SetRight(NilNode);
			fRootNode = newNode;
		} else
			return false;
	}
	
	newNode = 0;
	return true;
}

template <class T>
bool
splayTree<T>::Insert(const T & val)
{
	return Add(val);
}

template <class T>
bool
splayTree<T>::Remove(const T & val)
{
	binaryNode<T> *newTree;
	
	if (!Includes(val))
		return false;
		
	if (fRootNode->Left() == NilNode)
		newTree = fRootNode->Right();
	else
	{
		newTree = fRootNode->Left();
		Splay(val, newTree);
		newTree->SetRight(fRootNode->Right());
	}
	
	delete fRootNode;
	fRootNode = newTree;
	return true;
}

template <class T>
void
splayTree<T>::Splay(const T & val, binaryNode<T> * &node)
{
	binaryNode<T> *leftTreeMax;
	binaryNode<T> *rightTreeMin;
	binaryNode<T> header;
	
	header.SetLeft(NilNode);
	header.SetRight(NilNode);
	leftTreeMax = &header;
	rightTreeMin = &header;
	
	NilNode->fValue = val;
	
	for (;;)
	{
		if (val < node->fValue)
		{
			if (val < node->Left()->fValue)
				node = RotateWithLeftChild(node);
			if (node->Left() == NilNode)
				break;
			
			rightTreeMin->SetLeft(node);
			rightTreeMin = node;
			node = node->Left();
		} else if (node->fValue < val)
		{
			if (node->Right()->fValue < val)
				node = RotateWithRightChild(node);
			if (node->Right() == NilNode)
				break;
			
			leftTreeMax->SetRight(node);
			leftTreeMax = node;
			node = node->Right();
		} else
			break;
	}
	
	leftTreeMax->SetRight(node->Left());
	rightTreeMin->SetLeft(node->Right());
	node->SetLeft(header.Right());
	node->SetRight(header.Left());
}

template <class T>
void
splayTree<T>::FreeTree(binaryNode<T> *aTree)
{
	if (aTree != 0)
	{
		// Make sure we're not recursively trying to 
		// delete the same tree because either the left
		// or right nodes are the same as the tree itself.
		if ((aTree->Left() != aTree))
		{
			FreeTree(aTree->Left());
			aTree->SetLeft(0);
		}
		
		if ((aTree->Right() != aTree))
		{
			FreeTree(aTree->Right());
			aTree->SetRight(0);
		}
			
		// Now that the children have been deleted, delete the
		// node itself.
		if (aTree != NilNode)
			delete aTree;
	}
}

#endif
