#ifndef TREETRAVERSE_H
#define TREETRAVERSE_H

#include "iterator.h"
#include "binarynode.h"
#include "stacklist.h"

//===============================================================
// A binary tree is a tree made up of nodes, each of which has a
// value a left child, and a right child, which are also binary nodes.
// Traversing a binary tree can occur in six different orders.
//
// preorder		Process value, then left, then right
// inorder		Process left, the value, then right
// postorder	Process left, then right then value
// 				Process value, then right, then left
// 				Process right, then value, then left
// 				Process right, then left, then value
//
// The three that are named, preorder, inorder, postorder are the
// most widely used, and are implemented here.  If you find need
// for the others, they should be easily implemented by duplicating
// one of the existing ones.
//===============================================================

//===============================================================
// Template Interface: postorderTraversal
//
// This is a iterator that will do a postorder traversal of a binary
// tree.  If the tree looks like this:
//
//        B
//      /   \
//     A     C
//
// Then a postorder traversal will return the values in this order
// A C B
//===============================================================
template <class T>
class postorderTraversal : public iterator<T>
{
public:
				postorderTraversal(binaryNode<T> *root);
				
	// Implement the iterator protocol
	virtual	bool	init();
	virtual	long	next();
	virtual bool	done();
			
	virtual	T	operator()();
	virtual	long	operator++();
	virtual	void	operator=(T newValue);

	// Allow changing the root
			void	SetRoot(binaryNode<T> *newRoot);
		
protected:
	binaryNode<T>	*fRoot;
	stackList<binaryNode<T> *>	fIteratorStack;
	
			void	StackChildren(binaryNode<T> *);
			
private:
};

//---------------------------------------------------------------
//
//---------------------------------------------------------------
template <class T>
postorderTraversal<T>::postorderTraversal(binaryNode<T> *rootNode)
{
	fRoot = rootNode;
}

template <class T>
bool
postorderTraversal<T>::init()
{
	fIteratorStack.DeleteAll();
	StackChildren(fRoot);
	return (! fIteratorStack.IsEmpty());
}

template <class T>
void
postorderTraversal<T>::StackChildren(binaryNode<T> *node)
{
	fIteratorStack.Push(node);
	binaryNode<T> *next = node->Right();
	if (next)
		StackChildren(next);
	next = node->Left();
	if (next)
		StackChildren(next);
}

template <class T>
bool
postorderTraversal<T>::done()
{
	return fIteratorStack.IsEmpty();
}


template <class T>
long
postorderTraversal<T>::next()
{
	fIteratorStack.Pop();
	return (!fIteratorStack.IsEmpty());
}

template <class T>
T
postorderTraversal<T>::operator()()
{
	binaryNode<T> *node = fIteratorStack.Top();
	return node->fValue;
}

template <class T>
long
postorderTraversal<T>::operator ++ ()
{
	return next();
}

template <class T>
void
postorderTraversal<T>::operator = (T newValue)
{
	binaryNode<T> *node = fIteratorStack.Top();
	node->fValue = newValue;
}

//===============================================================
// Template Interface: preorderTraversal
//
// This is a iterator that will do a preorder traversal of a binary
// tree.  If the tree looks like this:
//
//        B
//      /   \
//     A     C
//
// Then a preorder traversal will return the values in this order
// B A C
//===============================================================
template <class T>
class preorderTraversal : public iterator<T>
{
public:
				preorderTraversal(binaryNode<T> *root);
				
	// Implement the iterator protocol
	virtual	bool	init();
	virtual	long	next();
	virtual bool	done();
			
	virtual	T		operator()();
	virtual	long	operator++();
	virtual	void	operator=(T newValue);

	// Allow changing the root
			void	SetRoot(binaryNode<T> *newRoot);
		
protected:
	binaryNode<T>	*fRoot;
	stackList<binaryNode<T> *>	fIteratorStack;
	
private:

};


//---------------------------------------------------------------
//
//---------------------------------------------------------------
template <class T>
preorderTraversal<T>::preorderTraversal(binaryNode<T> *rootNode)
{
	fRoot = rootNode;
}

template <class T>
bool
preorderTraversal<T>::init()
{
	fIteratorStack.DeleteAll();
	if (fRoot != 0)
		fIteratorStack.Push(fRoot);

	return (! fIteratorStack.IsEmpty());
}

template <class T>
bool
preorderTraversal<T>::done()
{
	return fIteratorStack.IsEmpty();
}


template <class T>
long
preorderTraversal<T>::next()
{
	binaryNode<T> *current = fIteratorStack.Top();
	binaryNode<T> *next = current->Left();
	
	if (next)
	{
		fIteratorStack.Push(next);
		return true;
	}
	
	while (!fIteratorStack.IsEmpty())
	{
		current = fIteratorStack.Pop();
		next = current->Right();
		if (next)
		{
			fIteratorStack.Push(next);
			return true;
		}	
	}
	
	return false;
}

template <class T>
T
preorderTraversal<T>::operator()()
{
	binaryNode<T> *node = fIteratorStack.Top();
	return node->fValue;
}

template <class T>
long
preorderTraversal<T>::operator ++ ()
{
	return next();
}

template <class T>
void
preorderTraversal<T>::operator = (T newValue)
{
	binaryNode<T> *node = fIteratorStack.Top();
	node->fValue = newValue;
}

//===============================================================
// Template Interface: inorderTraversal
//
// This is a iterator that will do a inorder traversal of a binary
// tree.  If the tree looks like this:
//
//        B
//      /   \
//     A     C
//
// Then a inorder traversal will return the values in this order
// A B C
//===============================================================
template <class T>
class inorderTraversal : public iterator<T>
{
public:
				inorderTraversal(binaryNode<T> *root);
				
	// Implement the iterator protocol
	virtual	bool	init();
	virtual	long	next();
	virtual bool	done();
			
	virtual	T		operator()();
	virtual	long	operator++();
	virtual	void	operator=(T newValue);

	// Allow changing the root
			void	SetRoot(binaryNode<T> *newRoot);
		
protected:
	binaryNode<T>	*fRoot;
	stackList<binaryNode<T> *>	fIteratorStack;
	
			void	SlideLeft(binaryNode<T> *);
			
private:


};


//---------------------------------------------------------------
//
//---------------------------------------------------------------
template <class T>
inorderTraversal<T>::inorderTraversal(binaryNode<T> *rootNode)
{
	fRoot = rootNode;
}

template <class T>
void
inorderTraversal<T>::SetRoot(binaryNode<T> *newRoot)
{
	fRoot = newRoot;
}

template <class T>
void
inorderTraversal<T>::SlideLeft(binaryNode<T> *current)
{
	while(current)
	{
		fIteratorStack.Push(current);
		current = current->Left();
	}
}

template <class T>
bool
inorderTraversal<T>::init()
{
	fIteratorStack.DeleteAll();
	SlideLeft(fRoot);
	
	return (! fIteratorStack.IsEmpty());
}

template <class T>
bool
inorderTraversal<T>::done()
{
	return fIteratorStack.IsEmpty();
}


template <class T>
long
inorderTraversal<T>::next()
{
	if (!fIteratorStack.IsEmpty())
	{
		binaryNode<T> *n = fIteratorStack.Pop();
		binaryNode<T> *next = n->Right();
		if (next)
			SlideLeft(next);
	}
	
	return (!fIteratorStack.IsEmpty());
}

template <class T>
T
inorderTraversal<T>::operator()()
{
	binaryNode<T> *node = fIteratorStack.Top();
	return node->fValue;
}

template <class T>
long
inorderTraversal<T>::operator ++ ()
{
	return next();
}

template <class T>
void
inorderTraversal<T>::operator = (T newValue)
{
	binaryNode<T> *node = fIteratorStack.Top();
	node->fValue = newValue;
}


#endif
